\name{scde-package}
\alias{scde-package}
\alias{scde}
\docType{package}
\title{
Since Cell Differential Expression (SCDE)
}
\description{
SCDE implements routines for Bayesian analysis of single-cell RNA-seq data.
}
\details{
\tabular{ll}{
Package: \tab scde\cr
Type: \tab Package\cr
Version: \tab 1.0\cr
Date: \tab 2013-09-27\cr
License: \tab GNU General Public License v2 (www.gnu.org/licenses/gpl-2.0.html‎)\cr
}
Please see http://pklab.med.harvard.edu/scde/ for additional info.
}
\author{
Peter Kharchenko
}
\references{
  The rationale behind the method is outlined in the following publication:
  Kharchenko PV, Silberstein L, Scadden DT "Bayesian approach to single-cell differential expression analysis." Nature Methods ...
}
\examples{
# load example dataset
# dataset available here: http://pklab.med.harvard.edu/peterk/scde/sample.data.zip
cd <- read.table("sample.data.txt",header=T)

# factor determining cell types
sg <- gsub("(MEF|ESC).*","\\1",colnames(cd))

# number of local process cores to use in processing
n.cores <- 10;

# cross-fit all pairs of cells within a given cell group (type), to determine which genes
# likely have not failed to be detected in a given cell measurement
# the three-component mixture model fit below can be very computationally intensive
# use the threshold.segmentation=T option below for a quick estimation of likely "drop-out"
# events based on a fixed RPM threshold
#lin.mod <- calculate.crossfit.models(cd,sg,verbose=1,n.cores=n.cores,old.cfm=NULL)
lin.mod <- calculate.crossfit.models(cd,sg,n.cores=n.cores,threshold.segmentation=T,save.plots=T)

# fit the error models for each cell
# for each cell of a given cell type, will look up genes that are in sufficient number of correlated components in pairwise comparisons for this group
# use the mean estimates to determine "expected expression values" for a subset of genes (for this cell type)
# min.nonfailed specifies the minimum number of cells in which the gene should be observed
# (i.e. not dropped out) in order for it to be used in the model fitting procedure.
# and then fit the NB-poisson mixture model for this cell
lin.ifm <- calculate.individual.models(cd,sg,lin.mod,min.nonfailed=5,n.cores=n.cores,save.plots=T);

# save basic models
save(lin.mod,lin.ifm,file="lin.model.RData")

# make per-cell list of error models
lin.ifm.flat <- unlist(lin.ifm,recursive=F); gc();
names(lin.ifm.flat) <- gsub("(.+?)\\.(.+)$","\\2",names(lin.ifm.flat),perl=T); # simplify names
  
  
# obtain adjusted expression magnitude values
lin.fpkm <- data.frame(do.call(cbind,lapply(sn(colnames(cd)),function(nam) get.fpm.estimates(lin.ifm.flat[[nam]],cd[,nam]))))
rownames(lin.fpkm) <- rownames(cd)
  
# esimate probability that a given observation is a drop-out event
lin.fail <- do.call(cbind,unlist(calculate.failure.p(cd,lin.ifm),recursive=F))
colnames(lin.fail) <- gsub("(.+?)\\.(.+)$","\\2",colnames(lin.fail),perl=T); # simplify names
rownames(lin.fail) <- rownames(cd)
  
# estimate expression magnitude prior
# sets up the expression magnitude grid that will be used in calculating all posterior distributions
# lenght.out determines the resolution of the expression magnitude grid on which
# prior and all posterior probability distributions will be evaluted
lin.prior <- estimate.signal.prior(lin.fpkm,lin.fail,show.plot=F,length.out=400); 
  
# differential expression test for a particular gene
# returns Z-score
pdf(file="Tdh.ES.MEF.test.pdf",width=4,height=7);
one.sided.test.id("Tdh","ESC","MEF",ifm=lin.ifm,dm=cd,prior=lin.prior,show.plots=T,bootstrap=T,n.samples=1e3,return.both=F)
dev.off()


# run differential expression test for all genes for all genes
  
# first we calculate posterior distributions of expression values for each individual cell,
# as well as joint posterior distributions for the defined cell groups (types).
  
# full posterior estimation, difference calculations
lin.pmat <- calculate.posterior.matrices(cd,lin.ifm,lin.prior)

# calculate joint posterior distributions for every cell type
# using 100 boostrap samples
lin.jpmat <- lapply(lin.pmat,calculate.joint.posterior.matrix,n.samples=100,n.cores=n.cores);

# test for differential expression between MEF and ES cell types
# calculate posterior of the expression fold ratio
lin.ESC.MEF.bdiffp <- calculate.ratio.posterior(lin.jpmat[["ESC"]],lin.jpmat[["MEF"]],lin.prior);

# calculate Z scores
lin.ESC.MEF.bdiffp.Z <- get.ratio.posterior.Z.score(lin.ESC.MEF.bdiffp)

# calculate MLE estimates of the log-fold change ratios
lin.ESC.MEF.bdiffp.M <- as.numeric(colnames(lin.ESC.MEF.bdiffp))[apply(lin.ESC.MEF.bdiffp,1,which.max)]

# show Z-scores for the top 10 genes upregulated in ESC compared to MEF
sort(lin.ESC.MEF.bdiffp.Z,decreasing=T)[1:10]

# show Z-scores for the top 10 genes upregulated in MEF compared to ESC
sort(lin.ESC.MEF.bdiffp.Z,decreasing=F)[1:10]

# simple report table
# determine MLE join expression level
expression.levels <- do.call(cbind,lapply(lin.jpmat,function(d) as.numeric(colnames(d))[apply(d,1,which.max)]))
report.table <- data.frame(gene=names(lin.ESC.MEF.bdiffp.Z),Z=lin.ESC.MEF.bdiffp.Z,M=lin.ESC.MEF.bdiffp.M,expression.levels)
report.table <- report.table[order(abs(report.table$Z),decreasing=T),]
head(report.table)
}
